import numpy as np
import matplotlib.pyplot as plt

class Mesh:
    def __init__(self, vertices=None, cells=None):
        if vertices is None:
            self.vertices = np.empty((0, 0), dtype=np.float64)
        else:
            self.vertices = np.array(vertices, dtype=np.float64)   # copy
            
        if cells is None:
            self.cells = np.empty((0, 0), dtype=np.int)
        else:
            self.cells = np.array(cells, dtype=np.int)   # copy
            
    def build_edges(self):
        # loops over the triangles and store the edges
        edges = np.empty((self.cells.shape[0] * 3, 2), dtype=np.int)
        for icell, cell in enumerate(self.cells):
            edges[3*icell] = cell[0:2]
            edges[3*icell + 1] = cell[1:3]
            edges[3*icell + 2] = cell[[2, 0]]
                      
        self.edges, counts = np.unique(np.sort(edges, axis=1), axis=0, return_counts=True)
        self.interior_edges = self.edges[counts == 2]
        self.boundary_edges = self.edges[counts == 1]
        
    def __repr__(self):
        plt.axis('equal')
        plt.plot(self.vertices[:, 0], self.vertices[:, 1], 'k.')
        ex = np.array([self.vertices[self.edges[:, 0], 0], self.vertices[self.edges[:, 1], 0]])
        ey = np.array([self.vertices[self.edges[:, 0], 1], self.vertices[self.edges[:, 1], 1]])
        plt.plot(ex, ey, 'k')
        plt.show()
        return f'Mesh with {self.cells.shape[0]} triangles and {self.vertices.shape[0]} vertices'
    
    def compute_areas(self):
        v1 = self.vertices[self.cells[:, 1]] - self.vertices[self.cells[:, 0]]
        v2 = self.vertices[self.cells[:, 2]] - self.vertices[self.cells[:, 0]]
        self.areas = 0.5 * np.abs(v1[:, 0] * v2[:, 1] - v2[:, 0] * v1[:, 1])
        
    def compute_normals(self):
        v = self.vertices[self.edges[:, 1]] - self.vertices[self.edges[:, 0]]
        norms = np.sqrt(v[:, 0]*v[:, 0] + v[:, 1]*v[:, 1])
        v /= norms[:, np.newaxis]
        self.normals = np.array([-v[:, 1], v[:, 0]]).T
        
    def read_mesh(self, filename):
        with open(filename) as f:
            while f.readline().strip() != "$Nodes":
                pass
            nnodes = np.int(f.readline())
            self.vertices = np.empty((nnodes, 2), dtype=np.float64)
            for inode in range(nnodes):
                line = f.readline().split()
                self.vertices[inode] = np.array([np.float64(line[1]), np.float64(line[2])])
                
            while f.readline().strip() != "$Elements":
                pass
            nelements = np.int(f.readline())
            self.cells = np.empty((nelements, 3), dtype=np.int)
            icell = 0
            for ielem in range(nelements):
                line = f.readline().split()
                if np.int(line[1]) == 2:
                    self.cells[icell] = np.array([np.int(line[-3])-1, np.int(line[-2])-1, np.int(line[-1])-1])
                    icell += 1
                
            self.cells.resize((icell, 3))
